@@ -1,11 +1,14 @@
 import argparse
+import functools
 import json
+import operator
 import os
 from collections.abc import MutableMapping
 from dataclasses import MISSING as _MISSING
 from dataclasses import Field, asdict, dataclass, fields, is_dataclass
+from pathlib import Path
 from pprint import pprint
-from typing import Any, Generic, List, TypeVar, Union, get_type_hints
+from typing import Any, Dict, Generic, List, Optional, Type, TypeVar, Union, get_type_hints
 
 T = TypeVar("T")
 MISSING: Any = "???"
@@ -19,18 +22,19 @@ NoDefaultVar = Union[_NoDefault[T], T]
 no_default: NoDefaultVar = _NoDefault()
 
 
-def is_common_type(arg_type: Any) -> bool:
-    """Check if the input type is one of the common types (int, float, str, bool).
+def is_primitive_type(arg_type: Any) -> bool:
+    """Check if the input type is one of `int, float, str, bool`.
 
     Args:
         arg_type (typing.Any): input type to check.
 
     Returns:
-        bool: True if input type is one of `int, float, str`.
+        bool: True if input type is one of `int, float, str, bool`.
     """
-    if is_list(arg_type):
+    try:
+        return isinstance(arg_type(), (int, float, str, bool))
+    except (AttributeError, TypeError):
         return False
-    return isinstance(arg_type(), (int, float, str, bool))
 
 
 def is_list(arg_type: Any) -> bool:
@@ -96,6 +100,12 @@ def safe_issubclass(cls, classinfo) -> bool:
         return r
 
 
+def _coqpit_json_default(obj: Any) -> Any:
+    if isinstance(obj, Path):
+        return str(obj)
+    raise TypeError(f"Can't encode object of type {type(obj).__name__}")
+
+
 def _default_value(x: Field):
     """Return the default value of the input Field.
 
@@ -162,10 +172,100 @@ def _serialize(x):
     return x
 
 
-def _deserialize(x, field_type):
-    """Pick the right desrialization for the given object and the corresponding field type.
+def _deserialize_dict(x: Dict) -> Dict:
+    """Deserialize dict.
 
-    TODO: Refactor with the factory pattern.
+    Args:
+        x (Dict): value to deserialized.
+
+    Returns:
+        Dict: deserialized dictionary.
+    """
+    out_dict = {}
+    for k, v in x.items():
+        if v is None:  # if {'key':None}
+            out_dict[k] = None
+        else:
+            out_dict[k] = _deserialize(v, type(v))
+    return out_dict
+
+
+def _deserialize_list(x: List, field_type: Type) -> List:
+    """Deserialize values for List typed fields.
+
+    Args:
+        x (List): value to be deserialized
+        field_type (Type): field type.
+
+    Raises:
+        ValueError: Coqpit does not support multi type-hinted lists.
+
+    Returns:
+        [List]: deserialized list.
+    """
+    field_args = None
+    if hasattr(field_type, "__args__") and field_type.__args__:
+        field_args = field_type.__args__
+    elif hasattr(field_type, "__parameters__") and field_type.__parameters__:
+        # bandaid for python 3.6
+        field_args = field_type.__parameters__
+    if field_args:
+        if len(field_args) > 1:
+            raise ValueError(" [!] Coqpit does not support multi-type hinted 'List'")
+        field_arg = field_args[0]
+        # if field type is TypeVar set the current type by the value's type.
+        if isinstance(field_arg, TypeVar):
+            field_arg = type(x)
+        return [_deserialize(xi, field_arg) for xi in x]
+    return x
+
+
+def _deserialize_union(x: Any, field_type: Type) -> Any:
+    """Deserialize values for Union typed fields
+
+    Args:
+        x (Any): value to be deserialized.
+        field_type (Type): field type.
+
+    Returns:
+        [Any]: desrialized value.
+    """
+    for arg in field_type.__args__:
+        # stop after first matching type in Union
+        try:
+            x = _deserialize(x, arg)
+            break
+        except ValueError:
+            pass
+    return x
+
+
+def _deserialize_primitive_types(x: Union[int, float, str, bool], field_type: Type) -> Union[int, float, str, bool]:
+    """Deserialize python primitive types (float, int, str, bool).
+    It handles `inf` values exclusively and keeps them float against int fields since int does not support inf values.
+
+    Args:
+        x (Union[int, float, str, bool]): value to be deserialized.
+        field_type (Type): field type.
+
+    Returns:
+        Union[int, float, str, bool]: deserialized value.
+    """
+
+    if isinstance(x, (str, bool)):
+        return x
+    if isinstance(x, (int, float)):
+        if x == float("inf") or x == float("-inf"):
+            # if value type is inf return regardless.
+            return x
+        x = field_type(x)
+        return x
+    # TODO: Raise an error when x does not match the types.
+    return None
+
+
+def _deserialize(x: Any, field_type: Any) -> Any:
+    """Pick the right desrialization for the given object and the corresponding field type.
 
     Args:
         x (object): object to be deserialized.
@@ -175,51 +275,43 @@ def _deserialize(x, field_type):
         object: deserialized object
 
     """
+    # pylint: disable=too-many-return-statements
     if is_dict(field_type):
-        out_dict = {}
-        for k, v in x.items():
-            if v is None:  # if {'key':None}
-                out_dict[k] = None
-            else:
-                out_dict[k] = _deserialize(v, type(v))
-        return out_dict
+        return _deserialize_dict(x)
     if is_list(field_type):
-        field_args = None
-        if hasattr(field_type, "__args__") and field_type.__args__:
-            field_args = field_type.__args__
-        elif hasattr(field_type, "__parameters__") and field_type.__parameters__:
-            # bandaid for python 3.6
-            field_args = field_type.__parameters__
-        if field_args:
-            if len(field_args) > 1:
-                raise ValueError(" [!] Coqpit does not support multi-type hinted 'List'")
-            field_arg = field_args[0]
-            # if field type is TypeVar set the current type by the value's type.
-            if isinstance(field_arg, TypeVar):
-                field_arg = type(x)
-            return [_deserialize(xi, field_arg) for xi in x]
-        return x
+        return _deserialize_list(x, field_type)
     if is_union(field_type):
-        for arg in field_type.__args__:
-            # stop after first matching type in Union
-            try:
-                x = _deserialize(x, arg)
-                break
-            except ValueError:
-                pass
-        return x
+        return _deserialize_union(x, field_type)
     if issubclass(field_type, Serializable):
-        return field_type().deserialize(x)
-    if is_common_type(field_type):
-        if isinstance(x, (str, bool)):
-            return x
-        if isinstance(x, (int, float)):
-            if x == float('inf') or x == float('-inf'):
-                # if value type is inf return regardless.
-                return x
-            return field_type(x)
+        return field_type.deserialize_immutable(x)
+    if is_primitive_type(field_type):
+        return _deserialize_primitive_types(x, field_type)
     raise ValueError(f" [!] '{type(x)}' value type of '{x}' does not match '{field_type}' field type.")
 
+
+# Recursive setattr (supports dotted attr names)
+def rsetattr(obj, attr, val):
+    def _setitem(obj, attr, val):
+        return operator.setitem(obj, int(attr), val)
+
+    pre, _, post = attr.rpartition(".")
+    setfunc = _setitem if post.isnumeric() else setattr
+
+    return setfunc(rgetattr(obj, pre) if pre else obj, post, val)
+
+
+# Recursive getattr (supports dotted attr names)
+def rgetattr(obj, attr, *args):
+    def _getitem(obj, attr):
+        return operator.getitem(obj, int(attr), *args)
+
+    def _getattr(obj, attr):
+        getfunc = _getitem if attr.isnumeric() else getattr
+        return getfunc(obj, attr, *args)
+
+    return functools.reduce(_getattr, [obj] + attr.split("."))
+
+
 @dataclass
 class Serializable:
     """Gives serialization ability to any inheriting dataclass."""
@@ -291,8 +383,10 @@ class Serializable:
         for field in fields(self):
             # if field.name == 'dataset_config':
             if field.name not in data:
-                init_kwargs[field.name] = vars(self)[field.name]
-                continue
+                if field.name in vars(self):
+                    init_kwargs[field.name] = vars(self)[field.name]
+                    continue
+                raise ValueError(f' [!] Missing required field "{field.name}"')
             value = data.get(field.name, _default_value(field))
             if value is None:
                 init_kwargs[field.name] = value
@@ -305,6 +399,34 @@ class Serializable:
             setattr(self, k, v)
         return self
 
+    @classmethod
+    def deserialize_immutable(cls, data: dict) -> "Serializable":
+        """Parse input dictionary and desrialize its fields to a dataclass.
+
+        Returns:
+            Newly created deserialized object.
+        """
+        if not isinstance(data, dict):
+            raise ValueError()
+        data = data.copy()
+        init_kwargs = {}
+        for field in fields(cls):
+            # if field.name == 'dataset_config':
+            if field.name not in data:
+                if field.name in vars(cls):
+                    init_kwargs[field.name] = vars(cls)[field.name]
+                    continue
+                raise ValueError(f' [!] Missing required field "{field.name}"')
+            value = data.get(field.name, _default_value(field))
+            if value is None:
+                init_kwargs[field.name] = value
+                continue
+            if value == MISSING:
+                raise ValueError("Deserialized with unknown value for {} in {}".format(field.name, cls.__name__))
+            value = _deserialize(value, field.type)
+            init_kwargs[field.name] = value
+        return cls(**init_kwargs)
+
 
 # ---------------------------------------------------------------------------- #
 #                        Argument Parsing from `argparse`                      #
@@ -324,38 +446,48 @@ def _init_argparse(
     field_name,
     field_type,
     field_value,
+    field_default_factory,
     field_help,
     arg_prefix="",
     help_prefix="",
+    relaxed_parser=False,
 ):
-    if field_value is None and not is_common_type(field_type):
-        # if the field type not int, flot or str, do not add it to the arguments
+    if field_value is None and not is_primitive_type(field_type) and not is_list(field_type):
+        # aggregate types (fields with a Coqpit subclass as type) are not supported without None
         return parser
     arg_prefix = field_name if arg_prefix == "" else f"{arg_prefix}.{field_name}"
     help_prefix = field_help if help_prefix == "" else f"{help_prefix} - {field_help}"
-    if isinstance(field_type, dict):  # pylint: disable=no-else-raise
-        # TODO: currently I don't need it
-        raise NotImplementedError(
-            " [!] Parsing `dict` field from argparse is not yet implemented. Please create an issue."
+    if is_dict(field_type):  # pylint: disable=no-else-raise
+        # NOTE: accept any string in json format as input to dict field.
+        parser.add_argument(
+            f"--{arg_prefix}",
+            dest=arg_prefix,
+            default=json.dumps(field_value) if field_value else None,
+            type=json.loads,
         )
     elif is_list(field_type):
         # TODO: We need a more clear help msg for lists.
-        if len(field_type.__args__) > 1:
+        if len(field_type.__args__) > 1 and not relaxed_parser:
             raise ValueError(" [!] Coqpit does not support multi-type hinted 'List'")
         list_field_type = field_type.__args__[0]
-        if field_value is None:  # pylint: disable=no-else-raise
-            # if the list is None, assume insertion of a new value/object to the list[0]
-            # parser = _init_argparse(parser,
-            #                         '0.',
-            #                         list_field_type,
-            #                         list_field_type(),
-            #                         field_help='',
-            #                         help_prefix=f'insert a new {list_field_type} to index 0 - {help_prefix} - ',
-            #                         arg_prefix=f'{arg_prefix}.')
-            # TODO: it complicates parsing argparse
-            raise NotImplementedError(" [!] Please create an issue.")
+
+        # TODO: handle list of lists
+        if is_list(list_field_type) and relaxed_parser:
+            return parser
+
+        if field_value is None or field_default_factory is list:
+            if not is_primitive_type(list_field_type) and not relaxed_parser:
+                raise NotImplementedError(" [!] Empty list with non primitive inner type is currently not supported.")
+
+            # If the list's default value is None, the user can specify the entire list by passing multiple parameters
+            parser.add_argument(
+                f"--{arg_prefix}",
+                nargs="*",
+                type=list_field_type,
+                help=f"Coqpit Field: {help_prefix}",
+            )
         else:
-            # if a list is defined, just enable editing the values from argparse
+            # If a default value is defined, just enable editing the values from argparse
             # TODO: allow inserting a new value/obj to the end of the list.
             for idx, fv in enumerate(field_value):
                 parser = _init_argparse(
@@ -363,26 +495,44 @@ def _init_argparse(
                     str(idx),
                     list_field_type,
                     fv,
+                    field_default_factory,
                     field_help="",
                     help_prefix=f"{help_prefix} - ",
                     arg_prefix=f"{arg_prefix}",
+                    relaxed_parser=relaxed_parser,
                 )
     elif is_union(field_type):
         # TODO: currently I don't know how to handle Union type on argparse
-        raise NotImplementedError(
-            " [!] Parsing `Union` field from argparse is not yet implemented. Please create an issue."
-        )
+        if not relaxed_parser:
+            raise NotImplementedError(
+                " [!] Parsing `Union` field from argparse is not yet implemented. Please create an issue."
+            )
     elif issubclass(field_type, Serializable):
-        return field_value.init_argparse(parser, arg_prefix=arg_prefix, help_prefix=help_prefix)
-    elif is_common_type(field_type):
+        return field_value.init_argparse(
+            parser, arg_prefix=arg_prefix, help_prefix=help_prefix, relaxed_parser=relaxed_parser
+        )
+    elif isinstance(field_type(), bool):
+
+        def parse_bool(x):
+            if x not in ("true", "false"):
+                raise ValueError(f' [!] Value for boolean field must be either "true" or "false". Got "{x}".')
+            return x == "true"
+
         parser.add_argument(
-            f"--coqpit.{arg_prefix}",
+            f"--{arg_prefix}",
+            type=parse_bool,
+            default=field_value,
+        )
+    elif is_primitive_type(field_type):
+        parser.add_argument(
+            f"--{arg_prefix}",
             default=field_value,
             type=field_type,
             help=f"Coqpit Field: {help_prefix}",
         )
     else:
-        raise NotImplementedError(f" [!] '{field_type}' is not supported by arg_parser. Please file a bug report.")
+        if not relaxed_parser:
+            raise NotImplementedError(f" [!] '{field_type}' is not supported by arg_parser. Please file a bug report.")
     return parser
 
 
@@ -520,9 +670,13 @@ class Coqpit(Serializable, MutableMapping):
     def from_dict(self, data: dict) -> None:
         self = self.deserialize(data)  # pylint: disable=self-cls-assignment
 
+    @classmethod
+    def new_from_dict(cls: Serializable, data: dict) -> "Coqpit":
+        return cls.deserialize_immutable(data)
+
     def to_json(self) -> str:
         """Returns a JSON string representation."""
-        return json.dumps(asdict(self), indent=4)
+        return json.dumps(asdict(self), indent=4, default=_coqpit_json_default)
 
     def save_json(self, file_name: str) -> None:
         """Save Coqpit to a json file.
@@ -550,70 +704,104 @@ class Coqpit(Serializable, MutableMapping):
         self = self.deserialize(dump_dict)  # pylint: disable=self-cls-assignment
         self.check_values()
 
-    def parse_args(self, args: argparse.Namespace) -> None:
+    def parse_args(
+        self, args: Optional[Union[argparse.Namespace, List[str]]] = None, arg_prefix: str = "coqpit"
+    ) -> None:
         """Update config values from argparse arguments with some meta-programming ✨.
 
         Args:
-            args (namespace): argeparse namespace as an output of ```argparse.parse_arguments()```.
-
-        Returns:
-            Coqpit: new config object with updated values.
+            args (namespace or list of str, optional): parsed argparse.Namespace or list of command line parameters. If unspecified will use a newly created parser with ```init_argparse()```.
+            arg_prefix: prefix to add to CLI parameters. Gets forwarded to ```init_argparse``` when ```args``` is not passed.
         """
-        if isinstance(args, argparse.Namespace):
-            args_dict = vars(args)
-        elif isinstance(args, list):
-            # overriding values from .parse_known_args()
-            args_dict = dict([*zip(args[::2], args[1::2])])
+        if not args:
+            # If args was not specified, parse from sys.argv
+            parser = self.init_argparse(arg_prefix=arg_prefix)
+            args = parser.parse_args()
+        if isinstance(args, list):
+            # If a list was passed in (eg. the second result of `parse_known_args`, run that through argparse first to get a parsed Namespace
+            parser = self.init_argparse(arg_prefix=arg_prefix)
+            args = parser.parse_args(args)
+
+        args_dict = vars(args)
+
         for k, v in args_dict.items():
-            if k.replace("--", "").startswith("coqpit") and "." in k:
-                _, k = k.split(".", 1)
-                names = k.split(".")
-                cmd = "self"
-                for name in names:
-                    if str.isnumeric(name):
-                        cmd += f"[{name}]"
-                    else:
-                        cmd += f".{name}"
-                try:
-                    exec(cmd)  # pylint: disable=exec-used
-                except (TypeError, AttributeError) as e:
-                    raise Exception(f" [!] '{k}' not exist to override from argparse.") from e
+            if k.startswith(f"{arg_prefix}."):
+                k = k[len(f"{arg_prefix}.") :]
+            try:
+                rgetattr(self, k)
+            except (TypeError, AttributeError) as e:
+                raise Exception(f" [!] '{k}' not exist to override from argparse.") from e
 
-                if v is None:
-                    cmd += "= None"
-                elif isinstance(v, str):
-                    cmd += f"= '{v}'"
-                else:
-                    cmd += f"= {v}"
+            rsetattr(self, k, v)
 
-                exec(cmd)  # pylint: disable=exec-used
         self.check_values()
 
-    def init_argparse(self, parser: argparse.ArgumentParser, arg_prefix="", help_prefix="") -> argparse.ArgumentParser:
+    def parse_known_args(
+        self,
+        args: Optional[Union[argparse.Namespace, List[str]]] = None,
+        arg_prefix: str = "coqpit",
+        relaxed_parser=False,
+    ) -> List[str]:
+        """Update config values from argparse arguments. Ignore unknown arguments.
+           This is analog to argparse.ArgumentParser.parse_known_args (vs parse_args).
+
+        Args:
+            args (namespace or list of str, optional): parsed argparse.Namespace or list of command line parameters. If unspecified will use a newly created parser with ```init_argparse()```.
+            arg_prefix: prefix to add to CLI parameters. Gets forwarded to ```init_argparse``` when ```args``` is not passed.
+            relaxed_parser (bool, optional): If True, do not force all the fields to have compatible types with the argparser. Defaults to False.
+
+        Returns:
+            List of unknown parameters.
+        """
+        if not args:
+            # If args was not specified, parse from sys.argv
+            parser = self.init_argparse(arg_prefix=arg_prefix, relaxed_parser=relaxed_parser)
+            args, unknown = parser.parse_known_args()
+        if isinstance(args, list):
+            # If a list was passed in (eg. the second result of `parse_known_args`, run that through argparse first to get a parsed Namespace
+            parser = self.init_argparse(arg_prefix=arg_prefix, relaxed_parser=relaxed_parser)
+            args, unknown = parser.parse_known_args(args)
+
+        self.parse_args(args)
+        return unknown
+
+    def init_argparse(
+        self,
+        parser: Optional[argparse.ArgumentParser] = None,
+        arg_prefix="coqpit",
+        help_prefix="",
+        relaxed_parser=False,
+    ) -> argparse.ArgumentParser:
         """Pass Coqpit fields as argparse arguments. This allows to edit values through command-line.
 
         Args:
-            parser (argparse.ArgumentParser): argparse.ArgumentParser instance.
-            arg_prefix (str, optional): Prefix to be used for the argument name. Defaults to ''.
+            parser (argparse.ArgumentParser, optional): argparse.ArgumentParser instance. If unspecified a new one will be created.
+            arg_prefix (str, optional): Prefix to be used for the argument name. Defaults to 'coqpit'.
             help_prefix (str, optional): Prefix to be used for the argument description. Defaults to ''.
+            relaxed_parser (bool, optional): If True, do not force all the fields to have compatible types with the argparser. Defaults to False.
 
         Returns:
             argparse.ArgumentParser: parser instance with the new arguments.
         """
+        if not parser:
+            parser = argparse.ArgumentParser()
         class_fields = fields(self)
         field_value = None
         for field in class_fields:
             field_value = vars(self)[field.name]
             field_type = field.type
+            field_default_factory = field.default_factory
             field_help = _get_help(field)
             _init_argparse(
                 parser,
                 field.name,
                 field_type,
                 field_value,
+                field_default_factory,
                 field_help,
                 arg_prefix,
                 help_prefix,
+                relaxed_parser,
             )
         return parser
 
@@ -672,7 +860,7 @@ def check_argument(
         ), f" [!] prequested fields {prerequest} for {name} are not defined."
     # check if the path exists
     if is_path:
-        assert os.path.exists(c[name]), " [!] {c[name]} not exist."
+        assert os.path.exists(c[name]), f' [!] path for {name} ("{c[name]}") does not exist.'
     # skip the rest if the alternative field is defined.
     if alternative in c.keys() and c[alternative] is not None:
         return
